
// Single processor.  ``kernel-range-based'' translation of OVERALL.

#include "ad++.h"
#include <sys/time.h>

const int N = 100 ;
const int NITER = 10000 ;
//const int N = 10 ;
//const int NITER = 100 ;

int wlo [] = {1, 1}, whi [] = {1, 1} ;  // ghost widths for `writeHalo'

int main(int argc, char** argv) {
  AdlibInit(argc, argv) ;

  Procs2 p(1, 1) ;
  ON(p) {
    BlockRange x(N, p.dim(0), LAY_EXTENDED, 1, 1),
               y(N, p.dim(1), LAY_EXTENDED, 1, 1) ;

    float h = 1.0 / (N - 1) ;

    // Initialization:
    //    sol = exact solution
    //    f   = forcing term
    //    u   = approximation solution

    Array2<float> u(x, y), f(x, y), sol(x, y) ;

    const float fac = 4.0 - 2.0 * (cos(h) + cos(3 * h)) ;

    Index i(x), j(y) ;
    OVERALL(i) {
      OVERALL(j) {
	sol(i, j) = sin(3 * (h * i) + h * j) ;
        if(i == 0 || i == N - 1 || j == 0 || j == N - 1)
	  u(i, j) = sol(i, j) ;
        else {
	  u(i, j) = 0.0 ;
	  f(i, j) = fac * sol(i, j) ;
        }
      } ALLOVER(j) ;
    } ALLOVER(i) ;

//    gprintf("initial u:\n") ;
//    aprintf("%7.3f%*N", u, N) ;

    // Main loop

    Range y1 = y.triplet(1, N - 2) ;

    clock() ;

    for(int iter = 0 ; iter < NITER ; iter++) {
      for(int parity = 0 ; parity < 2 ; parity++) {

//      writeHalo(u, wlo, whi) ;

	float* u_dat = u.dat() ;
        Stride u0 = u.str(0), u1 = u.str(1) ;
        const int u0_len = u0.len(), u1_len = u1.len() ;

	float* f_dat = f.dat() ;
        Stride f0 = f.str(0), f1 = f.str(1) ;
        const int f0_len = f0.len(), f1_len = f1.len() ;

        Dimension jd = y1.dim() ;
        //apgStack.push(&apg) ;
        //apg.restrict(jd) ;

        int j_count, j_sub_bas, j_sub_stp ;
        const int j_bas = y1.block(jd.crd(), &j_sub_bas, &j_count) ;
        const int j_stp = y1.block_step(&j_sub_stp) ;

        const int u1_bas = u1.disp(j_sub_bas) ;
        const int u1_stp = u1.disp_step(j_sub_stp) ;

        const int f1_bas = f1.disp(j_sub_bas) ;
        const int f1_stp = f1.disp_step(j_sub_stp) ;

        for(int jl = 0 ; jl < j_count ; jl++) {
          Range x1 = x.triplet(1 + (1 + j_bas + j_stp * jl +
                                    parity) % 2, N - 2, 2) ;

	  Dimension id = x1.dim() ;
	  //apgStack.push(&apg) ;
	  //apg.restrict(id) ;

	  int i_count, i_sub_bas, i_sub_stp ;
	  const int i_bas = x1.block(id.crd(), &i_sub_bas, &i_count) ;
	  const int i_stp = x1.block_step(&i_sub_stp) ;

	  const int u0_bas = u0.disp(i_sub_bas) ;
	  const int u0_stp = u0.disp_step(i_sub_stp) ;

	  const int f0_bas = f0.disp(i_sub_bas) ;
	  const int f0_stp = f0.disp_step(i_sub_stp) ;

	  for(int il = 0 ; il < i_count ; il++)
	    u_dat[u0_bas + u0_stp * il + u1_bas + u1_stp * jl] =
		0.25 * (f_dat[f0_bas + f0_stp * il +
			      f1_bas + f1_stp * jl] +
			u_dat[u0_bas + u0_stp * il + u0_len * 1 +
			      u1_bas + u1_stp * jl] +
			u_dat[u0_bas + u0_stp * il - u0_len * 1 +
			      u1_bas + u1_stp * jl] +
			u_dat[u0_bas + u0_stp * il +
			      u1_bas + u1_stp * jl + u1_len * 1] +
			u_dat[u0_bas + u0_stp * il +
			      u1_bas + u1_stp * jl - u1_len * 1]) ;

	  //apgStack.pop(&apg) ;
	}

        //apgStack.pop(&apg) ;
      }

//      gprintf("iteration = %d\n", iter) ;
//      aprintf("%7.3f%*N", u, N) ;
    }

    clock_t t = clock() ;
  
    printf("elapsed time = %f\n", (float) t / CLOCKS_PER_SEC) ;

//    gprintf("exact solution:\n") ;
//    aprintf("%7.3f%*N", sol, N) ;

  } NO(p) ;

  AdlibFinalize() ;
}

